from ._dataset import *
from ._types import *

DATA_SETS = dict()
for spectrum in SpectrumType:
    DATA_SETS.update(
        {
            spectrum: Dataset.create(spectrum)
        }
    )


class Material(object):
    def __init__(self, plastic_type: Plastic):
        if not isinstance(plastic_type, Plastic):
            raise ValueError(f'Invalid type of plastic: {plastic_type} \n '
                             f'Supported plastic types: {[e.value for e in Plastic]}')
        self._plastic_type = plastic_type
        self._spectrum = dict()

        for spectrum_type in SpectrumType:
            self._spectrum.update(
                {
                    spectrum_type: DATA_SETS[spectrum_type].get().loc[self._plastic_type.value].sample().iloc[0]
                }
            )

    def spectrum(self, spectrum_type: SpectrumType):
        return self._spectrum[spectrum_type]



# upper and lower bounds in cm
MIN_CONTAINER_SIZE = 5
MAX_CONTAINER_SIZE = 15

# init location in cm
INIT_CONTAINER_X = 0
INIT_CONTAINER_Y = 0
INIT_CONTAINER_Z = 0

# upper and lower bounds for the gap between generated containers in cm
MIN_CONTAINERS_GAP = 5
MAX_CONTAINERS_GAP = 20

SIMULATION_FREQUENCY_HZ = 10
VALID_SENSORS_FREQUENCIES_HZ = [10, 5, 2, 1]  # divisors of SIMULATION_FREQUENCY_HZ

SAMPLING_FREQUENCY_TO_SNR_DB = {freq: 50 // freq for freq in VALID_SENSORS_FREQUENCIES_HZ}

NOISE_MU_VALUE = 0
EPSILON_POWER = 0.0001



import numpy as np

from ._constants import *
from ._container import Container
from ._material import DATA_SETS
from ._types import SimulationMode
from ._types import SpectrumType


class Sensor:
    _num_sensors = 0

    def __init__(self, sensor_type: SpectrumType, location_cm: int, sensor_id: int = None):
        if location_cm < 0:
            raise ValueError(f'Sensor location should be >= 0.')
        if not isinstance(sensor_type, SpectrumType):
            raise ValueError(f'Invalid type of sensor: {SpectrumType} \n '
                             f'Supported sensors types: {[e.value for e in SpectrumType]}')

        self._location_cm = location_cm
        self._sensor_type = sensor_type
        self._background_spectrum = DATA_SETS[self._sensor_type].background

        if sensor_id is None:
            Sensor._num_sensors += 1
            self._id = Sensor._num_sensors
        else:
            self._id = sensor_id

    @classmethod
    def reset_num(cls):
        cls._num_sensors = 0

    @classmethod
    def create(cls, sensor_type: SpectrumType, location: int, sensor_id: int = None):
        return Sensor(sensor_type, location, sensor_id)

    @property
    def location(self):
        return self._location_cm

    @property
    def type(self):
        return self._sensor_type

    @property
    def id(self):
        return self._id

    def read(self, container: Container, mode: str, sampling_frequency: int):
        if container is None:
            raw_output = self._background_spectrum
        else:
            raw_output = container.material.spectrum(self._sensor_type)

        if mode == SimulationMode.Training.value:
            return raw_output
        elif mode == SimulationMode.Testing.value:
            signal_power = raw_output ** 2
            signal_average_power = np.mean(signal_power)
            signal_average_power_db = 10 * np.log10(signal_average_power + EPSILON_POWER)
            noise_average_power_db = signal_average_power_db - SAMPLING_FREQUENCY_TO_SNR_DB[sampling_frequency]
            noise_average_power = 10 ** (noise_average_power_db / 10)
            noise = np.absolute(np.random.normal(NOISE_MU_VALUE, np.sqrt(noise_average_power), raw_output.size))
            return (raw_output + noise).rename("unknown_plastic")
        else:
            raise ValueError(f'Invalid reading mode,\n'
                             f'valid options: {[mode.value for mode in SimulationMode]}')



import uuid

from ._material import *
from ._types import *


class Container:
    def __init__(self, plastic_type: Plastic, dimension: ContainerDimension, location: ContainerLocation):
        if not isinstance(plastic_type, Plastic):
            raise ValueError(f'Invalid type of plastic for a Container: {plastic_type} \n '
                             f'Supported plastic types: {[e.value for e in Plastic]}')

        if not isinstance(dimension, ContainerDimension):
            raise ValueError(f'Invalid type of dimension for a Container: {dimension}')

        self._plastic_type = plastic_type
        self._material = Material(plastic_type)
        self._dimension = dimension
        self._location = location
        self._guid = uuid.uuid4()

    @classmethod
    def create(cls, plastic_type: Plastic, length: int, width: int, height: int):
        return Container(
            plastic_type,
            ContainerDimension(length, width, height),
            ContainerLocation(0, 0, 0)
        )

    @property
    def material(self):
        return self._material

    @property
    def plastic_type(self):
        return self._plastic_type

    @property
    def dimension(self):
        return self._dimension

    @property
    def location(self):
        return self._location

    @property
    def guid(self):
        return self._guid.hex

    @location.setter
    def location(self, new_location: ContainerLocation):
        if not isinstance(new_location, ContainerLocation):
            raise ValueError(f'Invalid type of location for a Container: {new_location}')
        self._location = new_location



import random
import types
from typing import List

from ._constants import *
from ._container import *
from ._conveyor import *
from ._sensor import *
from ._types import SimulationMode


class RecyclingPlant:
    def __init__(
            self,
            sorting_function,
            num_containers: int,
            conveyor: Conveyor,
            sensors: List[Sensor],
            mode: str):
        if not isinstance(sorting_function, types.FunctionType):
            raise ValueError(f'Invalid sorting function, please pass a compatible function')

        if not any(mode == simulation_mode.value for simulation_mode in SimulationMode):
            raise ValueError(f'Invalid simulation mode , \n'
                             f'valid options: {[mode.value for mode in SimulationMode]}')

        if conveyor is None:
            raise ValueError(f'No conveyor is passed')

        if sensors is None:
            raise ValueError(f'No sensor list is passed')

        self._user_sorting_function = sorting_function
        self._mode = mode
        self._conveyor = conveyor
        self._sensors = sensors
        self._num_remaining_containers = num_containers

        self._containers_list = []

        if self._mode == 'training':
            random.seed(1)

    @staticmethod
    def _is_visible_to_sensor(sensor: Sensor, container: Container):
        return container.location.y > sensor.location > (container.location.y - container.dimension.length)

    def _generate_container(self):
        if not self._containers_list or\
                (self._containers_list[-1].location.y - self._containers_list[-1].dimension.length) > \
                random.randint(MIN_CONTAINERS_GAP, MAX_CONTAINERS_GAP):
            self._containers_list.append(
                Container(
                    random.choice(list(Plastic)[0:-1]),  # Added - not generate containers with blank spectrum
                    ContainerDimension(
                        random.randint(MIN_CONTAINER_SIZE, MAX_CONTAINER_SIZE),
                        random.randint(MIN_CONTAINER_SIZE, MAX_CONTAINER_SIZE),
                        random.randint(MIN_CONTAINER_SIZE, MAX_CONTAINER_SIZE),
                    ),
                    ContainerLocation(
                        INIT_CONTAINER_X,
                        INIT_CONTAINER_Y,
                        INIT_CONTAINER_Z
                    )
                )
            )
            self._num_remaining_containers -= 1

    def update(
            self,
            current_iteration: int,
            simulation_frequency_hz: int,
            sensors_frequency_hz: int):

        missed = 0
        classified = 0
        mistyped = 0
        sensors_output = {}
        sensed_containers = {}
        identification_result = {}

        # init sensor_output with the background spectrum
        for sensor in self._sensors:
            sensors_output.update(
                {
                    sensor.id: {
                        'type': sensor.type,
                        'location': sensor.location,
                        'spectrum': sensor.read(None, self._mode, sensors_frequency_hz),
                    }
                }
            )

        time_step_sec = 1 / simulation_frequency_hz
        for container in self._containers_list:
            container.location.y += time_step_sec * self._conveyor.speed
            if container.location.y >= self._conveyor.dimension.length:
                self._containers_list.remove(container)
                missed += 1  # container reached the end of the conveyor
            for sensor in self._sensors:
                if self._is_visible_to_sensor(sensor, container):
                    sensors_output.update(
                        {
                            sensor.id: {
                                'type': sensor.type,
                                'location': sensor.location,
                                'spectrum': sensor.read(container, self._mode, sensors_frequency_hz),
                            }
                        }
                    )
                    sensed_containers.update(
                        {
                            sensor.id: container
                        }
                    )

        if self._num_remaining_containers != 0:
            self._generate_container()

        identification_output = None
        if current_iteration % (simulation_frequency_hz // sensors_frequency_hz) == 0:
            identification_output = self._user_sorting_function(sensors_output)

        if identification_output is not None:
            for sensor_id, plastic_type in identification_output.items():
                if plastic_type != Plastic.Blank:  # Only if it's not background, then compare
                    if sensor_id in sensed_containers.keys():
                        identification_result.update(
                            {
                                sensed_containers[sensor_id].guid: {
                                    'actual_type': sensed_containers[sensor_id].plastic_type.value,
                                    'identified_type': plastic_type.value,
                                }
                            }
                        )
                        if sensed_containers[sensor_id].plastic_type == plastic_type:
                            classified += 1
                        else:
                            mistyped += 1
                        self._containers_list.remove(sensed_containers[sensor_id])

        is_done = self._num_remaining_containers == 0 and len(self._containers_list) == 0
        return missed, classified, mistyped, identification_result, is_done



from ._constants import *
from ._recycling_plant import *
from ._types import SimulationMode


class RPSimulation:
    def __init__(
            self,
            sorting_function,
            num_containers: int,
            sensors: List[Sensor],
            sampling_frequency: int,
            conveyor: Conveyor,
            mode
    ):
        if not any(mode == simulation_mode.value for simulation_mode in SimulationMode):
            raise ValueError(f'Invalid simulation mode,\n'
                             f'valid options: {[mode.value for mode in SimulationMode]}')

        if sampling_frequency not in VALID_SENSORS_FREQUENCIES_HZ:
            raise ValueError(f'Invalid sampling frequency,\n'
                             f'valid options: {[frequency for frequency in VALID_FREQUENCIES]}')

        self._simulation_frequency_hz = SIMULATION_FREQUENCY_HZ
        self._sensors_frequency_hz = sampling_frequency

        self._sorting_function = sorting_function
        self._num_containers = num_containers
        self._conveyor = conveyor
        self._sensors = sensors
        self._mode = mode

        self._recycling_plant = None

        self._total_missed = 0
        self._total_classified = 0
        self._total_mistyped = 0
        self._current_iteration = 0
        self._identification_result = {}

        self._init_recycling_plant()
        
        # to start from 1 for the next simulation
        Sensor.reset_num()

    def _init_recycling_plant(self):
        self._recycling_plant = RecyclingPlant(
            self._sorting_function,
            self._num_containers,
            self._conveyor,
            self._sensors,
            self._mode)

    @property
    def total_missed(self):
        return self._total_missed

    @property
    def total_classified(self):
        return self._total_classified

    @property
    def total_mistyped(self):
        return self._total_mistyped

    @property
    def identification_result(self):
        return self._identification_result

    def reset(self):
        self._init_recycling_plant()

        self._total_missed = 0
        self._total_classified = 0
        self._total_mistyped = 0
        self._current_iteration = 0
        self._identification_result = {}

    def _update(self):
        missed, classified, mistyped, identification_result, is_done = self._recycling_plant.update(
            self._current_iteration,
            self._simulation_frequency_hz,
            self._sensors_frequency_hz
        )
        self._current_iteration += 1
        self._total_missed += missed
        self._total_classified += classified
        self._total_mistyped += mistyped
        self._identification_result.update(identification_result)

        return is_done

    def run(self):
        self.reset()
        while not self._update():
            pass
        return self._current_iteration * (1 / self._simulation_frequency_hz)



import uuid

from ._types import ConveyorDimension


class Conveyor:
    def __init__(self, speed_cm_per_second: int, dimension: ConveyorDimension):
        if speed_cm_per_second < 0:
            raise ValueError(f'Invalid speed: {speed_cm_per_second}'
                             f'Conveyor speed should be >= 0 cm per second.')

        if not isinstance(dimension, ConveyorDimension):
            raise ValueError(f'Invalid type of dimension for a conveyor: {dimension}')

        self._speed_cm_per_second = speed_cm_per_second
        self._dimension = dimension
        self._guid = uuid.uuid4()

    @classmethod
    def create(cls, speed_cm_per_second: int, length: int, width: int):
        return Conveyor(
            speed_cm_per_second,
            ConveyorDimension(length, width)
        )

    @property
    def speed(self):
        return self._speed_cm_per_second

    @property
    def dimension(self):
        return self._dimension

    @property
    def guid(self):
        return self._guid.hex

    @speed.setter
    def speed(self, speed_cm_per_second: int):
        if speed_cm_per_second < 0:
            raise ValueError(f'Invalid speed: {speed_cm_per_second} \n'
                             f'Conveyor speed should be >= 0 cm per second.')

        self._speed_cm_per_second = speed_cm_per_second



import enum


class SpectrumType(enum.Enum):
    FTIR = 'FTIR'
    Raman = 'Raman'


class Plastic(enum.Enum):
    HDPE = 'HDPE'
    LDPE = 'LDPE'
    PP = 'PP'
    PS = 'PS'
    PC = 'PC'
    PVC = 'PVC'
    Polyester = 'Polyester'
    PET = 'PET'
    PU = 'PU'
    Blank = 'background'


class SimulationMode(enum.Enum):
    Testing = 'testing'
    Training = 'training'


class ContainerLocation:
    def __init__(self, x, y, z):
        self._x = x
        self._y = y
        self._z = z

    @property
    def x(self):
        return self._x

    @property
    def y(self):
        return self._y

    @property
    def z(self):
        return self._z

    @x.setter
    def x(self, new_x):
        self._x = new_x

    @y.setter
    def y(self, new_y):
        self._y = new_y

    @z.setter
    def z(self, new_z):
        self._x = new_z


class ContainerDimension:
    def __init__(self, length, width, height):
        if length <= 0:
            raise ValueError(f'Invalid value for length: {length}; should be > 0.')
        if width <= 0:
            raise ValueError(f'Invalid value for width: {width}; should be > 0.')
        if height <= 0:
            raise ValueError(f'Invalid value for height: {height}; should be > 0.')

        self._length = length
        self._width = width
        self._height = height

    @property
    def length(self):
        return self._length

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @length.setter
    def length(self, new_length):
        if new_length <= 0:
            raise ValueError(f'Invalid value for length: {new_length}; should be > 0.')
        self._length = new_length

    @width.setter
    def width(self, new_width):
        if new_width <= 0:
            raise ValueError(f'Invalid value for width: {new_width}; should be > 0.')
        self._width = new_width

    @height.setter
    def height(self, new_height):
        if new_height <= 0:
            raise ValueError(f'Invalid value for height: {new_height}; should be > 0.')
        self._height = new_height


class ConveyorDimension:
    def __init__(self, length, width):
        if length <= 0:
            raise ValueError(f'Invalid value for length: {length}; should be > 0.')
        if width <= 0:
            raise ValueError(f'Invalid value for width: {width}; should be > 0.')
        self._length = length
        self._width = width

    @property
    def length(self):
        return self._length

    @property
    def width(self):
        return self._width

    @length.setter
    def length(self, new_length):
        if new_length <= 0:
            raise ValueError(f'Invalid value for length: {new_length}; should be > 0.')
        self._length = new_length

    @width.setter
    def width(self, new_width):
        if new_width <= 0:
            raise ValueError(f'Invalid value for width: {new_width}; should be > 0.')
        self._width = new_width



from setuptools import setup, find_packages

setup(
    name='rcplant',
    version='1.1.3',
    url='',
    license='',
    author='Mohammadreza Rostam, Rongxuan Du',
    author_email='reza.rostam@mech.ubc.ca',
    description='Recycling Plant Simulator',
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: Free for non-commercial use",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering"
    ],
    packages=find_packages(),
    include_package_data=True,
    package_data={'rcplant/data': ['rcplant/data/*.xlsx']},
    python_requires='>=3.7',
    install_requires=["pandas>=1.3.4", "openpyxl>=3.0.9", "numpy>=1.21.5"]
)



import os

import pandas as pd

from ._types import SpectrumType


class Dataset(object):

    def background(self):
        pass

    def get(self):
        pass

    @classmethod
    def create(cls, dataset_type):
        if dataset_type == SpectrumType.FTIR:
            return FTIR()
        elif dataset_type == SpectrumType.Raman:
            return Raman()
        else:
            raise ValueError(f'${dataset_type} is not a valid dataset type, \n'
                             f'Supported dateset types: {[e.value for e in SpectrumType]}')


class FTIR(Dataset):
    def __init__(self):
        data_folder = os.path.join(os.path.dirname(__file__), 'data')
        data_file = os.path.join(data_folder, 'FTIR Plastic Database.xlsx')
        self._data_table = pd.read_excel(data_file, sheet_name=0, index_col=0)
        # it's assumed that the background spectrum is all zeros for now, should be replaced with the actual data
        self._background = self._data_table.loc['background']

    def get(self):
        return self._data_table

    @property
    def background(self):
        return self._background

    def __str__(self):
        return 'FTIR'


class Raman(Dataset):
    def __init__(self):
        data_folder = os.path.join(os.path.dirname(__file__), 'data')
        data_file = os.path.join(data_folder, 'Raman Plastic Database.xlsx')
        self._data_table = pd.read_excel(data_file, sheet_name=0, index_col=0)
        # it's assumed that the background spectrum is all zeros for now, should be replaced with the actual data
        self._background = self._data_table.loc['background']

    def get(self):
        return self._data_table

    @property
    def background(self):
        return self._background

    def __str__(self):
        return 'Raman'
